#!/usr/bin/perl -w
#**********************************************************************
#
# Utility to administer virtual domains in postfix (both the standard
# ones and those defined in http://jth.net/virtual.html)
#
# This program should reside in /var/www/perl/virtdomain.pl with global
# read and execute rights (chmod o+rx /var/www/perl/virtdomain.pl).
#
# The first argument to the program determines its function:

# CREATE:  call from commandline as root: create virtual domain by updating
#          /etc/postfix/virtual and virtual_regexp
#          and creating /etc/virtual/domain/passwd and /var/spool/virtual/domain
#          CREATE <domain name> <real user administering the domain>
#                               <passwd> [postfix]
#          <real user administering the domain> will be the postmaster of the domain

# DELETE:  call from commandline as root: delete virtual domain by updating
#          /etc/postfix/virtual and virtual_regexp
#          and deleting /etc/virtual/domain/passwd and /var/spool/virtual/domain as
#          well as .forward files in the administrators homedir
#          DELETE <domain name> <real user administering the domain> <passwd>

# <blank>
# MAILBOUNCE
#          if called from postfix as mail to user@domain the email on STDIN will be processed
#          according to the setup of the user in /etc/virtual/domain/passwd.
#          MAILBOUNCE: bounce email to unknown users otherwise store in postmaster mailbox

# <blank>  if called from postfix as mail to postmaster@domain or <postfixadmin+postmaster>
#          with the subject "*** Virtual domain update ***" and line 1 is trans=UPDATE....
#          a user will be created/updated/deleted
#          otherwise process email normally as above

# <blank>: if called from a web server displays webpages to update the user
#          base in a virtual domain e,g, http://jth.net/perl/virtdomain.pl

#
# As a basis the user postmaster must be defined for the domain.
# in the /etc/virtual/domain/passwd file. The user and group ids as
# well as homedir from this user is used for the virtual domain.
# For each virtual user the file homedir/.forward+user--domain is created.
# The web server should have read access to these files as well as
# the /etc/virtual/domain/passwd file.
# The GNU pop3d deamon as maintained by me is supporting this
# virtual domain implementation.
#
# Using this program to deliver the mail for jth style domains is not suitable for
# a high volume system, as loading the Perl interpreter
# has a fairly high initial CPU usage compared to this program (90% of total run time).
# However, using the provision for hooking in another program to deliver the mail
# may be useful. An awk utility will be released shortly after this release.
#
# 2002 Jrgen Thomsen jth@jth.net, http://jth.net/virtual.html
# 1.01  2002-08-22 Official release
# 1.02  2003-01-24 Support for new virtual_alias_map parameter and relay_domains behaviour
# 1.03  2003-04-26 Better language support and fix webmail URL in info
# 1.04  2003-05-30 Chars | and " were removed from forwards in input. Now user shall not enter |, but start commands with "
# 1.05  2004-08-06 Modifications specific to jth.net: /var/www/perl, body background changed, no CGI used
#
#**********************************************************************
$main::postfixadmin = '';  # user allowed to update the postfix virtual_xxx_maps files and /etc/virtual/domain/passwd
$main::mydomain  = '';        # Mail with the updates is mailed to 'postmaster@domain' (jth domains) or
                              # $main::postfixemail below for all postfix style domains. Sender is
                              # <virt admin user>@$main::admindomain (jth style domains)
                              # $main::postfixadmin.'@'.$main::mydomain (postfix domains)
                              # A forward+postmaster file is created in the postfixadmin homedir to call
                              # this program for updating the postfix files running under the proper user id
$main::masterpw     = '0500'; # set to a value if a master password for all domains is wanted
# the following is for displayUserHelp only
$main::POP3server   = 'pop3'; # terminate by '.' if fixed for all, else displayed as "$main::POP3server.$domain"
$main::SMTPserver   = 'smtp'; # as POP3 server, if empty specified as 'your ISP SMTP-server'
$main::webmail      = 'mail'; # as POP3 server, if empty not displayed else http://domain/$main::webmail
$main::showuidgid   = 'yes';  # yes: display and change user id and group id, no: use standard values for all users
$main::showmailbox  = 'yes';  # yes: display and change user mailbox, no: use standard value for all users
$main::vm_pop3d     = 'no';   # yes: your POP3 server is VM-pop3d (encrypted passwords, reduced functions)

# see html_init for other values to change

$main::progversion     = '1.05 2004-08-06'; # program version
use strict;
use warnings;
use User::pwent;
use Sys::Syslog qw(:DEFAULT setlogsock);
our $content_type = "Content-type: text/html\n\n";
setlogsock("unix");
openlog("virtdomain", "pid", "mail");
my ($res, @users, @pairs, $buffer, %Form);
$main::progname = substr($0, rindex($0,"/")+1); # filename without path eg. virtdomain.pl
($main::progid) = split(/\./, $main::progname); # filenames without extension eg. virtdomain

$res = 0;
my ($uid,$gid,$name,$homedir, @forwards, @newforwards);

$main::domainstyle = 'jth';  # default; jth: use jth style virtual domains, postfix: use postfix style virtual domains
                             # determined for a domain by testing existence of /home/user/.forward+postmaster--domain
$main::command = '';
if (defined($ARGV[4]) && lc($ARGV[4]) eq 'postfix') {$main::domainstyle = 'postfix'};
$main::command = "WEB" if exists $ENV{REQUEST_METHOD};
$main::command = "MAIL" if exists $ENV{LOCAL};
$main::command = uc($ARGV[0]) if defined($ARGV[0]) && uc($ARGV[0]) !~ m!(/|-)(H|\?)!;

undef @users;
undef @main::newforwards;
undef $main::Pvuser;
undef $main::Pname;
undef %Form;
%Form = ();
$buffer = '';

html_init();

$Form{'trans'} = '';
$Form{'dom'} = '';
$Form{'pwd'} = '';
$Form{'action'} = '';

if ($main::command =~ /^(CREATE|DELETE)$/) {
    $Form{'trans'} = $main::command;
    $Form{'dom'}   = $ARGV[1];
    $Form{'user'}  = $ARGV[2];
    $Form{'pwd'}   = $ARGV[3];
} else {

    if ($main::command =~ /^MAIL/) { # called from Postfix
        readHeader();
        if (($main::subject && $main::subject eq '*** Virtual domain update ***') &&
            (exists $ENV{EXTENSION}) && ($ENV{EXTENSION} =~ /^(postmaster--|postmaster$)/)) {
            $buffer = <STDIN>;
            $buffer =~ s/^\s+//;
            $buffer =~ s/\s+$//;
        }
    } elsif ($main::command =~ /^WEB/) {
       if ($ENV{'REQUEST_METHOD'}eq 'GET') {
           $buffer = $ENV{'QUERY_STRING'};
       }
       elsif ($ENV{'REQUEST_METHOD'} eq 'POST') {
           read(STDIN, $buffer, $ENV{'CONTENT_LENGTH'});
       }
    }
#    my $tmp = $buffer; $tmp =~ s/%/!/g;
#    syslog('info', $main::command.': '.$tmp);

    @pairs = split(/&/, $buffer);

    # For each name-value pair:
    foreach my $pair (@pairs) {

        # Split the pair up into individual variables.
        my ($name, $value) = split(/=/, $pair); $value = '' if !$value;

        # Decode the form encoding on the name and value variables
        $name =~ tr/+/ /;
        $name =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;

        $value =~ tr/+/ /;
        $value =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;

        # erase server side includes
        $value =~ s/<!--(.|\n)*-->//g;
        $value =~ s/^\s+//; $value =~ s/\s+$//;
        $value =~ s/[\>|<*&!?%\\]+/ /g; #get rid of strange characters (security risk)
        $Form{$name} = $value;
        push @main::newforwards, $value if $name =~ /^fwd/ && $value ne '';
    }
    if ($main::command =~ /^WEB/) {
#        use CGI qw(-compile :all);
#        my $q = CGI->new();
#        print $q->header();
         print $content_type;
    }
}

$Form{'dom'} = lc($Form{'dom'}); $Form{'dom'} =~ s/[\|><*&!?]+//g;
$main::domain = $Form{'dom'};
$main::lang = $Form{'lang'} if exists $Form{'lang'};
@main::t = texts($main::lang);
$main::command = 'WEBLOGIN' if !$Form{'trans'} && $main::command eq 'WEB';
$main::command = 'WEBCHECK' if $Form{'trans'} && $Form{'trans'} =~ /^$main::t[49]/i;
$main::command = 'WEBUHELP' if $Form{'trans'} && $Form{'trans'} =~ /^$main::t[48]/i;

$main::pagetitle = "Virtual domain check" if $main::command eq 'WEBCHECK';
$main::pagetitle = "Virtual domain login" if $main::command eq 'WEBLOGIN';

$main::res = 0;
$main::res = program_init($main::command); #  OBS. check this procedure for proper configuration parameters for your system

if ($main::res) {
    htmlPrint("$main::res");
    syslog('info', $main::res);
    htmlFooter();
    exit 75;
}

if (!$main::command || ($main::command !~ /^(CREATE|DELETE|WEB|WEBLOGIN|WEBUHELP|MAIL|MAILBOUNCE)$/) ||
   ($main::command =~ /CREATE|DELETE/ && (! defined($ARGV[1]) || ! defined($ARGV[2]) || ! defined($ARGV[3])))) {
    htmlPrint("\n<p>$main::t[32]:<br />");
    htmlPrint("Syntax: <br />");
    htmlPrint("   $main::progname CREATE &lt;domain&gt; &lt;admin user/postmaster&gt; &lt;virt.dom passwd&gt; [postfix|jth]<br />");
    htmlPrint("   $main::progname DELETE &lt;domain&gt; &lt;admin user/postmaster&gt; &lt;virt.dom. passwd&gt;</p>");
    htmlPrint('<p><a href="'.$ENV{'SCRIPT_NAME'}.'">Login ('.$ENV{'SCRIPT_NAME'}.')</a></p>') if $main::command =~ /^WEB/;
    syslog('info', "Syntax error: COMMAND = $main::command"); closelog();
    htmlFooter();
    exit 75; # temporary failure
}

if (!$main::postfixadmin && ($main::domainstyle eq 'postfix')) {
    htmlPrint("Enter user with proper update rights to postfix files in \$main::postfixadmin<br />");;
    syslog('info', '$main::postfixadmin not defined');
    htmlFooter();
    exit 75;
}


my $pwdok = 1;

$main::delivery =~ s/DOMAIN/$Form{'dom'}/;
#syslog('info', $ENV{EXTENSION}." $main::domainstyle, $Form{'mailbox'}");
$Form{'mailbox'} = '.' if !defined($Form{'mailbox'}) ||
        ($Form{'mailbox'} eq "$main::t[0]") ||
        ($Form{'mailbox'} eq "$main::virtual_mailbox_base/$Form{'dom'}/$Form{'user'}") ||
        (($main::domainstyle eq 'jth') && ($Form{'mailbox'} !~ m!^/!)) ||
        (($main::domainstyle eq 'postfix') && ($Form{'mailbox'} =~ m!^/!));
#syslog('info', $ENV{EXTENSION}." $Form{'mailbox'}");

if ($main::command =~ /^MAIL/) { # either updates from the web page or normal incoming mail
    if ($Form{'trans'} ne 'UPDATE') { # normal mail
        $res = storeMessage($main::command);
    } else {
        $pwdok = getUsers($Form{'dom'}, $Form{'pwd'}, $Form{'user'}, 'new');
        if ($pwdok) {
            updateUser($Form{'dom'}, $Form{'pwd'}, $Form{'user'}, $main::domainstyle, $Form{'upwd'},$Form{'uid'},
                $Form{'gid'},$Form{'name'}, $Form{'mailbox'}, $Form{'action'}, $Form{'bounce'},
                'nokeepold', $Form{'encrypt'}, $Form{'forwardonly'}, @main::newforwards);
        }
    }
} elsif ($main::command =~ '^WEB') { # display of web pages
    if ($main::command eq 'WEBUHELP') {
        displayUserHelp();
    } elsif ($main::command eq 'WEBLOGIN' || $Form{'dom'} eq '' || ! -e $main::virtual_conf_base.'/'.$Form{'dom'}.'/passwd') {
        displayLoginPage();
    } elsif ($main::command ne 'WEBCHECK') {
        $pwdok = getUsers($Form{'dom'}, $Form{'pwd'}, $Form{'user'}, 'new');
#syslog("info", "%s", $Form{'dom'}.", ".$Form{'pwd'}.", ".$Form{'user'});
        if (!$pwdok) {
            htmlPrint('<center>'.$pwdok.$main::t[18].'</center>'); # wrong password
        } elsif (($Form{'trans'} eq '') || ($Form{'trans'} eq 'LIST')) {
            listUsers($Form{'dom'}, $Form{'pwd'}, $main::domainstyle);
        } elsif ($Form{'trans'} eq 'LIST1') {
            listOneUser($Form{'dom'}, $Form{'pwd'}, $Form{'user'}, $main::domainstyle);
        } elsif ($Form{'trans'} eq 'UPDATE') {
            emailUpdateUser($Form{'dom'}, $Form{'pwd'}, $Form{'user'}, $main::domainstyle, $buffer);
        } else {
            htmlPrint('<center>'.$main::t[16].'</center>'); # unknown command
        }
    }
} elsif ($main::command =~ /^(CREATE|DELETE)$/) { # command line for create/delete virtual domains
    domains($Form{'trans'}, $Form{'dom'}, $Form{'user'}, $Form{'pwd'}, $main::domainstyle);
}

htmlFooter();
closelog();

exit $res;

sub html_init {
# *****************************************
# HTML values for web pages
#$main::banner = "<center><h1><b>jth virtual domains</b><h1></center>"; images/overtop.jpg
#$main::banner = "<center><h1><a href=\"http://jth.net\"><img alt=\"jth.net\" src=\"http://jth.net/icons/jthbanner.gif\" border=\"0\" /></a></h1></center>";
# OBS: Danmark, Grnland og Frerne:
# Uden skriftlig tilladelse af Jrgen Thomsen, jth@jth.net m $main::banner ikke ndres eller forhindres vist !

$main::bg         = "#FFFFFF"; # background colour
$main::bg         = "url(\"/icons/sandgrey.gif\")"; # image  e.g. "url(\"/images/bg.gif\")"
$main::fg         = "#000000"; # text color

$main::link       = "#ff0000"; # link color
$main::visited    = "#30b000"; # visited link color
$main::formmethod = 'post'; # post or get, only get for debugging
$main::pagetitle  = "Virtual domain administration";
$main::domain     = '';
$main::headerwritten = 0;

# determine language on web page (see subroutine texts)
$main::lang = '';
if (exists $ENV{HTTP_ACCEPT_LANGUAGE}) {
    my @langs = split(/,/, $ENV{HTTP_ACCEPT_LANGUAGE});
    foreach my $l (@langs) {
        $main::lang = 'en' if $l =~ /en/ && $main::lang eq ''; # English found
        $main::lang = 'da' if $l =~ /da/ && $main::lang eq ''; # Danish  found
    }
}
if (!$main::lang && exists $ENV{'LANG'}) {
    $main::lang = $ENV{'LANG'};
    $main::lang =~ s/^([^_]+).*$/$1/;
}
}

sub program_init {
# *****************************************
# get posfix configuration values and set other values
my ($cmd) = @_;
my $tmp;

undef $main::bcc;
#$main::bcc = 'virtdom@jth.net';                    # uncomment and change email address for getting copies of all updates
$main::webserveruser = 'apache';                    # user under which your web server is running

$main::virtual_mailbox_base = '/var/spool/virtual'; # where the virtual mailbox files usually reside in dir/domain
$main::virtual_conf_base    = '/etc/virtual';       # where the virtual mailbox passwd files reside in dir/domain

$main::virtualproc  = "$0";                         # command to be stored in .forward files to call this program
$main::delivery = "awk  -F: -f /usr/local/bin/$main::progid.awk MAILBOUNCE=0 PHASE=1 /etc/virtual/DOMAIN/passwd PHASE=2 -";
                                                    # to be stored in .forward files to deliver mail to mailbox (high performance)
$main::delivery     = "$0";                         # to be stored in .forward files to deliver mail to mailbox (low performance)
$main::nextform     = "http://jth.net/index.da.php?scroll=yes&amp;ramme=/perl/$main::progname";         # how to call this program on a web page
$main::lockprogram  = '/usr/bin/lockfile -l 60';    # program to create a mailbox.lock file to prevent simultaneous accesses
$main::sync         = 'sync';                       # program to perform a fsync (flushing buffers to disk)
$main::postmap      = '/usr/sbin/postmap';          # location of postfix postmap program
$main::allowcmd     = 1;                            # allow commands in the .forward files: 0/1. OBS. admin can insert manually

if ($cmd =~ /^(WEBCHECK)?$/) {
    htmlPrint("\n$main::t[31]:<br />");
    htmlPrint('<table align="center">');
}
$main::mydomain     = postfix_get_var('mydomain') if $main::mydomain eq '';
if ($main::mydomain eq '') {
    $main::mydomain = postfix_get_var('myhostname');
    $main::mydomain =~ s/^[^.]*\.(.*)$/$1/;
}
$main::domain = $main::mydomain if !$main::domain;

$main::postfixemail = '';
$main::postfixemail = $main::postfixadmin.'+postmaster@'.$main::mydomain if $main::postfixadmin;

#$main::relay_domains        = postfix_get_var('relay_domains', 'all', '$mydestination');
$tmp                        = postfix_get_var('virtual_mailbox_base', '', '/var/spool/virtual');
#$main::virtual_alias_domains= postfix_get_var('virtual_alias_domains', 'all', '$virtual_alias_maps, $virtual_mailbox_maps');
$main::virtual_mailbox_base = $tmp if $tmp ne '';
$main::virtual_mailbox_maps = postfix_get_var('virtual_mailbox_maps', '', 'hash:/etc/postfix/vmailbox');
$main::virtual_uid_maps     = postfix_get_var('virtual_uid_maps', '', 'hash:/etc/postfix/vuid');
$main::virtual_gid_maps     = postfix_get_var('virtual_gid_maps', '', 'hash:/etc/postfix/vgid');
$main::transport_maps       = postfix_get_var('transport_maps', '', 'hash:/etc/postfix/transport');
$main::home_mailbox         = postfix_get_var('home_mailbox','noprint', $main::t[37]);
$main::mail_spool_directory = postfix_get_var('mail_spool_directory', '', '/var/mail');
$main::sendmail_path        = postfix_get_var('sendmail_path', 'noprint', '/usr/sbin/sendmail');
$tmp = '';
if ($main::sendmail_path eq '') {
    $main::sendmail_path    = '/usr/sbin/sendmail';
    if (! -e $main::sendmail_path) {
        $tmp = '** OBS ** '; $main::sendmail_path = '';
        $main::res = $tmp."sendmail_path ".$main::t[38];
    }
}
htmlPrint("<tr><td>$tmp"."main.cf: sendmail_path </td><td>=</td><td> $main::sendmail_path</td></tr>")
    if ($cmd =~ /^(WEBCHECK)?$/);

$main::virtual_alias_maps = postfix_get_var('virtual_alias_maps', 'all', 'hash:/etc/postfix/virtual,regexp:/etc/postfix/virtual_regexp');
$main::virtual_regexp_maps = $main::virtual_alias_maps;
$tmp = $main::virtual_alias_maps;
$main::virtual_alias_maps =~ s/,/ /g;
$main::virtual_alias_maps =~ s!.*hash:([^ ]+).*$!$1!;
if ($tmp eq $main::virtual_alias_maps) {
    htmlPrint("<tr><td>** OBS ** Postfix virtual map file </td><td>=</td><td></td><td>[ /etc/postfix/virtual ]</td></tr>")
        if ($cmd =~ /^(WEBCHECK)?$/);
    $main::virtual_alias_maps = '';
} else {
    htmlPrint("<tr><td>Postfix virtual map file </td><td>=</td><td>$main::virtual_alias_maps</td></tr>")
        if ($cmd =~ /^(WEBCHECK)?$/);
}
$tmp = $main::virtual_regexp_maps;
$main::virtual_regexp_maps =~ s!.*(pcre:|regexp:)([^ ]+).*$!$2!;
if ($tmp eq $main::virtual_regexp_maps) {
    htmlPrint("<tr><td>** OBS ** Postfix virtual regexp map file </td><td>=</td><td></td><td>[ /etc/postfix/virtual_regexp ]</td></tr>")
        if ($cmd =~ /^(WEBCHECK)?$/);

    $main::virtual_regexp_maps = '';
} else {
    htmlPrint("<tr><td>Postfix virtual regexp map file </td><td>=</td><td>$main::virtual_regexp_maps</td></tr>")
        if ($cmd =~ /^(WEBCHECK)?$/);
}

#$main::virtual_regexp_maps  = "/etc/postfix/virtual_regexp";

$main::mailprog  = $main::sendmail_path.' -t ';      # program to send an email from the webserver
$main::usermail = 'mail';
$main::usermail = '** does not exist **' if !getpwnam($main::usermail);

if ($cmd =~ /^(WEBCHECK)?$/) {
    htmlPrint("<tr><td>Local mail user (vm_pop3d)           </td><td>=</td><td> $main::usermail</td></tr>");;
    htmlPrint("<tr><td>GNU pop3d/jth configuration path     </td><td>=</td><td> $main::virtual_conf_base</td></tr>");;
    htmlPrint("<tr><td>Postfix virt. dom. admin       </td><td>=</td><td> $main::postfixadmin</td></tr>");
    htmlPrint("<tr><td>Postfix virt. dom. admin email </td><td>=</td><td> $main::postfixemail</td></tr>");
}

#if (($main::relay_domains !~ /virtual_alias_maps/) || ($main::relay_domains !~ /virtual_mailbox_maps/) ||
    if (!$main::virtual_alias_maps || !$main::virtual_regexp_maps || !$main::virtual_mailbox_base) {
    htmlPrint("<tr><td colspan=\"4\"><br />ERROR: /etc/postfix/main.cf $main::t[36]:</td></tr>");
#    htmlPrint('<br /><tr><td>relay_domains </td><td>=</td><td> $mydestination, $virtual_alias_maps, $virtual_mailbox_maps</td></tr>')
#       if ($main::relay_domains !~ /virtual_alias_maps/) || ($main::relay_domains !~ /virtual_mailbox_maps/);

    htmlPrint("<br /><tr><td>virtual_alias_maps </td><td>=</td><td> hash:/etc/postfix/virtual, regexp:/etc/postfix/virtual_regexp</td></tr>")
        if (!$main::virtual_alias_maps || !$main::virtual_regexp_maps);

    htmlPrint("<br /><tr><td>virtual_mailbox_base</td><td>=</td><td> /var/spool/virtual</td></tr>")
          if !$main::virtual_mailbox_base;

    $main::res = "<br /><b>/etc/postfix/main.cf $main::t[35] $main::progname</b>";
}
if (($cmd =~ /^(WEBCHECK)?$/) &&
   (!$main::virtual_mailbox_maps || !$main::virtual_uid_maps || !$main::virtual_gid_maps || !$main::transport_maps)) {
    htmlPrint("\n<tr><td colspan=\"4\"><br />$main::t[39]:<br />");
    htmlPrint("\ntouch $main::virtual_mailbox_maps<br />");
    htmlPrint("\npostmap $main::virtual_mailbox_maps</td></tr>");
}

htmlPrint("</table>") if ($cmd =~ /^(WEBCHECK)?$/);
return $main::res;
}

sub readHeader {
# *****************************************
# input: email on STDIN
# output: @main::headlines, $main::envsender, $main::subject,
#         email positioned on first line in body

my $lin = <STDIN>;
my $linno = 0;
my $dummy;
undef @main::headlines;

if ($lin =~ /^From /) {
    ($dummy, $main::envsender, $dummy) = split(/\s+/, $lin);
} else {
    $main::envsender = 'postbackup@jth.net';
}
chomp $lin;
push @main::headlines, $lin; $linno = 1;

while (($lin ne "") && ($lin ne "\r") && ($lin = <STDIN>)) {
    chomp $lin;
    ($dummy, $main::subject) = split(/:/, $lin) if $lin =~ /^Subject:/i;
    ($dummy, $main::origto) = split(/:/, $lin) if $lin =~ /^X-Original-To:/i;
    push @main::headlines, $lin; $linno++;
}
$main::subject =~ s/^\s+// if $main::subject;
$main::origto =~ s/^\s+//;

#syslog('info', "readHeader: lines=$linno") if $debug > 0;
}

sub storeMessage {
# *****************************************
# This procedure is only called for jth style virtual domains
# input: MAIL|MAILBOUNCE, @main::headlines, $main::envsender, email positioned on first line in body
# output: email stored in mailbox or exit 67 (bounce msg)

my ($cmd) = @_;
my $res = 0;
my $res1 = 0;
my $lin;
my $linno = 0;
my $hlinno = 0;
my $lastlin;
my $tmp;
my ($vdomain, $vuser,$upwd,$uid,$gid,$name,$homedir, $mailbox);

my $pop3user = $ENV{'USER'} if $ENV{'USER'};

if ( exists $ENV{'EXTENSION'} && ($ENV{'EXTENSION'} =~ /--/)) { # virtual domain user
    ($vuser, $vdomain) = split(/--/, $ENV{'EXTENSION'});
    $res = getUsers($vdomain, "", $vuser);
    my $ll = $vuser.':'; undef $tmp;
    foreach my $l (@main::users) {
        ($tmp,$upwd,$uid,$gid,$name,$mailbox) = split(/:/,$l) if $l =~ /^$ll/;
    }

    if (defined($tmp)) { # user found
        if (defined($mailbox) && ($mailbox eq '.')) {$mailbox = "$main::virtual_mailbox_base/$vdomain/$vuser"};
    } elsif ($cmd eq 'MAILBOUNCE') {
            syslog('info', "$vdomain: bounced: $pop3user, $vuser\@$vdomain from $main::envsender");
            exit 67; # bounce message
        }
}
if ((!defined($mailbox) || ($vuser eq 'postmaster')) && defined($pop3user)) { # unknown virtual user, store in admin mailbox
    my $pw = getpwnam($pop3user);
    my $maildir = $pw->dir."/$main::home_mailbox";
    $maildir =~ s!(^.*)/[^/]*$!$1!;
    if (($main::home_mailbox ne '') && -d $maildir) {
        $mailbox = $pw->dir."/$main::home_mailbox";
    } elsif (-d $main::mail_spool_directory) {
        $mailbox = $main::mail_spool_directory."/$pop3user";
    } elsif (-d "/var/mail") {
        $mailbox = "/var/mail/$pop3user";
    }
    if (!defined($vuser)) {$vuser = $ENV{'LOCAL'}; $vdomain = $ENV{'DOMAIN'};}
} elsif (!defined($pop3user)) {
    syslog('info', "$vdomain: ERROR: could not find environment variable USER. Something quite wrong");
    exit 67; # bounce msg
}

# lock mailbox
$res = 0; $res1 = 1;
$lastlin = '';
system("$main::lockprogram $mailbox.lock");
open MB, ">>".$mailbox or $res = 8;
syslog('info', "$vdomain: open: RC=$res: $mailbox $!") if $res != 0;

$main::dt = 0;
foreach my $hl (@main::headlines) {
    $hl = fixHeader($hl, $vuser, $vdomain, $main::envsender);
    $res1 = print MB $hl."\n";
    if (!$res1) { $res = 5; $res1 = $! }
}
syslog('info', "$vdomain: printH: RC=$res $mailbox $res1") if $res != 0;

while (($res == 0) && ($lin = <STDIN>)) {
    chomp $lin; $linno++;
    $lastlin = $lin;
    $lin = '>'.$lin if $lin =~ /^From /;
    $res1 = print MB $lin."\n";
    if (!$res1) { $res = 5; syslog('info', "$vdomain: printB: RC=$res $mailbox $!"); }
}
$res1 = print MB $lin."\n" if $lastlin !~ /^(\r|)$/;
if (!$res1) { $res = 6; syslog('info', "$vdomain: printB: RC=$res $mailbox $!"); }

$res1 = close MB;
if (!$res1) { $res = 7; syslog('info', "$vdomain: close: RC=$res $mailbox $!"); }
system($main::sync);

# unlock mailbox
#system("/bin/rm -f $mailbox.lock");
unlink "$mailbox.lock";

#
$tmp = "";
$tmp .= " Fejl=$res" if $res != 0;
$linno += $#main::headlines + 2;

syslog('info', "$vdomain: stored: $pop3user, $vuser\@$vdomain from $main::envsender in $mailbox, lines=$linno $tmp");
$res = 75 if $res != 0; # temporary failure
return $res;
}

sub fixHeader {
# *****************************************
# Change the Delivered-To line and add X-Envelope-To line
#
my ($lin, $vuser, $vdomain, $envsender) = @_;
if (($main::dt == 0) && ($lin =~ /Delivered-To:/i)) {
    $lin = "Delivered-To: $vuser\@$vdomain\n";
    $lin .= "X-Envelope-To: $vuser\@$vdomain";
    $main::dt = 1;
}
return $lin;
}

sub domains {
# *****************************************
# Create and delete virtual domains.
# Must be run as root
# Input: CREATE|DELETE, <virtual domain name>,
#        <user name> login name of user administering the domain (must exist)
#        <password> for administering the domain, should not be the admin system password
#        <style> postfix|jth, wanted domain style
#
my ($command, $domain, $user, $pwd, $style) = @_;
return if !defined($domain) || $command !~ /^(CREATE|DELETE)$/;

print "Style of virtual domain $domain specified as: $style\n";

if (($style eq 'postfix') &&
   (($main::virtual_mailbox_maps eq '') || ( $main::virtual_uid_maps eq '' ) || ($main::virtual_gid_maps eq '')
      || ($main::transport_maps eq ''))) {
   print "Insert values for virtual_mailbox_maps, virtual_uid_maps, virtual_gid_maps\n";
   print "and transport_maps in /etc/postfix/main.cf e.g.\n";
   print "virtual_mailbox_maps = hash:/etc/postfix/vmailbox\n";
   print "virtual_uid_maps = hash:/etc/postfix/vuid\n";
   print "virtual_gid_maps = hash:/etc/postfix/vgid\n";
   print "transport_maps = hash:/etc/postfix/transport\n";
   print "virtual_minimum_uid = 100\n";
#   print "Make the files writable for the postfix admin user:\n";
#   print "chown <admin user>.<his group> <filename>\nchmod 0644 <filename>\n";
   return;
} else {
   if (! -e $main::virtual_regexp_maps) {
       print "ERROR: File for Postfix virtual regexp not found: regexp:$main::virtual_regexp_maps";
       return;
   }
}

my $pw = getpwnam($main::webserveruser) or die "No user $main::webserveruser. Change \$main::webserveruser to your webserver user";
my $apachegid = $pw->gid if $pw;
$pw = getpwnam($main::usermail) or die "No user $main::usermail";
my $mailgid = $pw->gid if $pw;
my $tmp;

if ($style eq 'postfix') {
    $pw = getpwnam($main::postfixadmin) or die "No user $user";
} else {
    $pw = getpwnam($user) or die "No user $user";
}
$apachegid = $pw->gid if !$apachegid;
$mailgid = $pw->gid if !$mailgid;


#print $pw->name.":".$pw->passwd.":".$pw->uid.":".$pw->gid.":".$pw->gecos.":".$pw->dir.":".$pw->shell."\n";

if ($command eq 'CREATE') {
    domainCreation($domain, $user, $pwd, $style, $pw, $apachegid, $mailgid);
} elsif ($command eq 'DELETE') {
    domainDeletion($domain, $user, $pwd, $style, $pw);
}

if ($style eq 'postfix') {
    if (! -e $main::virtual_mailbox_maps) {system('touch '.$main::virtual_mailbox_maps);}
    chown $pw->uid, 0, $main::virtual_mailbox_maps;
    chmod 0664, $main::virtual_mailbox_maps;
    if (! -e $main::virtual_uid_maps) {system('touch '.$main::virtual_uid_maps);}
    chown $pw->uid, 0, $main::virtual_uid_maps;
    chmod 0664, $main::virtual_uid_maps;
    if (! -e $main::virtual_gid_maps) {system('touch '.$main::virtual_gid_maps);}
    chown $pw->uid, 0, $main::virtual_gid_maps;
    chmod 0664, $main::virtual_gid_maps;
#    print "Refreshing db $main::virtual_mailbox_maps\n";
    system($main::postmap." -o $main::virtual_mailbox_maps") == 0 or die "ERROR: postmap $main::virtual_mailbox_maps: $?";
#    print "Refreshing db $main::virtual_uid_maps\n";
    system($main::postmap." -o $main::virtual_uid_maps") == 0 or die "ERROR: postmap $main::virtual_uid_maps: $?";
#    print "Refreshing db $main::virtual_gid_maps\n";
    system($main::postmap." -o $main::virtual_gid_maps") == 0 or die "ERROR: postmap $main::virtual_gid_maps: $?";
#    chown $pw->uid, $pw->gid, $main::virtual_alias_maps;
#    print "Refreshing db $main::transport_maps\n";
    system($main::postmap." -o $main::transport_maps") == 0 or die "ERROR: postmap $main::transport_maps: $?";
}

}

sub domainConversion {
# *****************************************
# To be called before ordinary domain creation to handle existing data
#
my ($domain, $user, $pwd, $style, $pw, $oldstyle) = @_;
my $tmp;
my $bounce = '';
my $forward = '';
my $domain1 = $domain; $domain1 =~ s/\./\\\./g;
my $upwd;

print "Trying to convert $domain from '$oldstyle' to '$style' style virtual domain\n";

if ($style eq 'jth' || # convert from postfix style
    (($oldstyle eq $style) && ($style eq 'postfix'))) { # create/restore passwd file for postfix domain

    # get all the mailboxes from the virtual_mailbox_maps file and insert each into the /etc/virtual/passwd file
    $tmp = 'grep -E "^[^@]+@'.$domain1.'\s+" '.$main::virtual_mailbox_maps;
    my @mbs = `$tmp`;
    foreach my $m (@mbs) {
        $m =~ s/\s+$//;
        my ($key, $mailbox) = split(/\s+/, $m);
        my $uid = $main::postmap.' -q '.$key." $main::virtual_uid_maps"; $uid = `$uid`; $uid = s/\s+//;
        $uid = $pw->uid if $uid eq '';
        my $gid = $main::postmap.' -q '.$key." $main::virtual_gid_maps"; $gid = `$gid`; $gid = s/\s+//;
        $gid = $pw->gid if $gid eq '';
        my ($vuser) = split(/@/, $key);
        $mailbox = '.' if ($mailbox eq "$domain/$vuser");
        $upwd = 'password';

        # get the users forwarding options in virtual
        my $forward = $main::postmap.' -q '."$vuser\@$domain $main::virtual_alias_maps";
        $forward = `$forward`; $forward =~ s/\s+/ /;
        @main::forwards = split(/\s+/,$forward);
        $forward = 'no';
        if (@main::forwards && $main::forwards[0] eq "$vuser\@$domain") {
            shift @main::forwards;
        } elsif (@main::forwards) {$forward = ''}

        # create the .forward+user--domain file and update the passwd entry
        updateUser($domain, $pwd, $vuser, $style, $upwd, $uid, $gid, $vuser, $mailbox, 'update',
                   'nobounce', 'keepold', 'noencrypt', $forward, @main::forwards);
    }
    # set flag for the forward+postmaster--domain file for bouncing mail for unknown users
    $bounce = $main::postmap.' -q @'.$domain." $main::virtual_alias_maps"; $bounce = `$bounce`; $bounce =~ s/\s+$//g;
    $bounce = ($bounce eq '' ? 'yes' : 'no');

    # perform standard deletion but keeping passwd and mailboxes
    domainDeletion($domain, $user, $pwd, 'postfix', $pw, 'keepmail');

} else { # convert from jth style

    # create users in virtual_mailbox_maps, virtual_uid_maps, virtual_gid_maps from /etc/virtual/passwd
    open PWD, "<$main::virtual_conf_base/$domain/passwd";
    while (my $lin = <PWD>) {
        $lin =~ s/\s+$//;
        my ($vuser,$upwd,$uid,$gid,$name,$mailbox) = split(/:/, $lin);

        if ($mailbox =~ m!^/! && $mailbox !~ m!^$main::virtual_mailbox_base/$domain/$vuser!) { # copy non-standard mailbox to standard one
            if (-e $mailbox ) {
                system("cp -a $mailbox $main::virtual_mailbox_base/$domain/$vuser") == 0 or die "Mailbox not copied: $mailbox: $!";
            }
        }
        $mailbox = '.'; # also if mailbox = local username
        $bounce = '';
        if ($vuser eq 'postmaster') {
            $bounce = 'fgrep MAILBOUNCE '.pw->dir.'/.forward'; $bounce = `$bounce`; $bounce =~ s/\s+$//g;
            $bounce = ($bounce ne '' ? 'yes' : 'no');
        }

        $tmp = $pw->dir."/.forward+$user--$domain";
        $forward = 'no';
        if (-e $tmp) {
            my @lines = `cat $tmp`;
            @main::forwards = grep(m!/$main::progid|/virtual(1)?!, @lines);
            $forward = '' if @main::forwards;
            @main::forwards = grep(!m"$main::progid|\||/", @lines);
        }

        updateUser($domain, $pwd, $vuser, $style, $upwd, $uid, $gid, $vuser, $mailbox, 'update',
                    $bounce, 'nokeepold', 'noencrypt', $forward, @main::forwards);
    }
    close PWD;

    # delete .forward+user--domain files
    system('rm '.$pw->dir."/.forward+*--$domain") == 0 or die "Cannot delete .forward files: $!";

    # perform standard deletion but keeping passwd and mailboxes
    domainDeletion($domain, $user, $pwd, 'jth', $pw, 'keepmail');
}
return $bounce;
}

sub domainCreation {
# *****************************************
my ($domain, $user, $pwd, $style, $pw, $apachegid, $mailgid) = @_;
my ($tmp, $res);
my $bounce = ''; # don't bounce if unknown user
    # create /etc/virtual/domain/passwd with postmaster
    if (!(-d $main::virtual_conf_base)) { mkdir $main::virtual_conf_base, 0751;}
                                   else { chmod 0751, $main::virtual_conf_base;}
    if (!(-d $main::virtual_conf_base."/$domain")) { mkdir $main::virtual_conf_base."/$domain", 0751;}
    chown $pw->uid, $pw->gid, $main::virtual_conf_base."/$domain";

    chmod 0711, glob ($main::virtual_conf_base."/*") or die "Cannot set X-bit on $main::virtual_conf_base/*: $!";

    updateFile($domain, $main::virtual_conf_base."/$domain/passwd",
        "postmaster:".$pwd.":".$pw->uid.":".$pw->gid.":".$user.":".$pw->dir.":", "^postmaster:");
    chown $pw->uid, $apachegid, $main::virtual_conf_base."/$domain/passwd";
    chmod 0640, $main::virtual_conf_base."/$domain/passwd";

    # create mailbox spool dir /var/spool/virtual/domain
    if (!(-d $main::virtual_mailbox_base)) { mkdir $main::virtual_mailbox_base, 0755;}
    if (!(-d $main::virtual_mailbox_base."/$domain")) { mkdir $main::virtual_mailbox_base."/$domain", 0730;}
    chown $pw->uid, $mailgid, $main::virtual_mailbox_base."/$domain";
    chmod 0730, $main::virtual_mailbox_base."/$domain";
    my $domain1 = $domain; $domain1 =~ s/\./\\\./g;
    chmod 0664, $main::virtual_alias_maps;
    chmod 0664, $main::virtual_regexp_maps;


    # check if we are converting/refreshing an existing domain
    my $oldstyle = 'jth' if -e $pw->dir."/.forward+postmaster--$domain";
    if (!$oldstyle) { # check for existing postfix style cirtual domain
        $oldstyle = $main::postmap.' -q '.$domain." $main::virtual_mailbox_maps";
        $oldstyle = `$oldstyle`; $oldstyle =~ s/\s+$//;
        $oldstyle = 'postfix' if $oldstyle ne '';
    }

    if ($oldstyle ne '') { # convert/refresh existing virtual domain
        domainConversion($domain, $user, $pwd, $style, $pw, $oldstyle);
    }

    if ($style eq 'jth') { # jth style virtual domains

         # create /home/user/.forward+postmaster--domain
        $tmp = $main::virtualproc; $tmp =~ s/(\.|\|)/\\$1/g;
        updateFile($domain, $pw->dir."/.forward+postmaster--$domain", '|"'.$main::virtualproc.$bounce.'"', $tmp);
        chown $pw->uid, $pw->gid, $pw->dir."/.forward+postmaster--$domain";
        chmod 0644, $pw->dir."/.forward+postmaster--$domain";

        # create /home/user/.forward
        updateFile($domain, $pw->dir.'.forward', '|'.$main::virtualproc, $tmp);
        chown $pw->uid, $pw->gid, $pw->dir."/.forward";
        chmod 0644, $pw->dir.'/.forward';

        # allow webserver access to users homedir
        chown $pw->uid, $apachegid, $pw->dir;
        chmod 0710, $pw->dir;

        # insert into /etc/postfix/virtual
        updateFile($domain, $main::virtual_alias_maps, "$domain virtual", "^$domain1\\s+");

        # insert into /etc/postfix/virtual_regexp
        $tmp = $domain; $tmp =~ s/\./\\\\\\\./g;
        updateFile($domain, $main::virtual_regexp_maps, '/^(.+)@(.+\.)?('.$domain1.')$/ '.$user.'+$1--${2}${3}', $tmp);

    } else { # postfix style virtual domains

        chown $pw->uid, 0, $main::virtual_alias_maps;
        chown $pw->uid, 0, $main::virtual_regexp_maps;

        # insert into /etc/postfix/transport
        updateFile($domain, $main::transport_maps, "$domain virtual", "^$domain1\\s+");

        # insert into /etc/postfix/virtual
        updateFile($domain, $main::virtual_alias_maps, "postmaster\@$domain $user\@$main::mydomain", "^postmaster\@$domain1\\s+");
        updateFile($domain, $main::virtual_alias_maps, "\@$domain postmaster\@$domain", "^\@$domain1\\s+");

        # insert into /etc/postfix/vmailbox file
        updateFile($domain, $main::virtual_mailbox_maps, "$domain virtual", "^$domain1\\s+");

        #create .forward+postmaster file like the .forward+postmaster--domain file for domain updates from web
        $pw = getpwnam($main::postfixadmin) or die "No user $user";

        $tmp = $main::virtualproc; $tmp =~ s/(\.|\|)/\\$1/g;
        updateFile($domain, $pw->dir."/.forward+postmaster", '|"'.$main::virtualproc.'"', $tmp);
        chown $pw->uid, $pw->gid, $pw->dir."/.forward+postmaster";
        chmod 0644, $pw->dir."/.forward+postmaster";
    }
    $res = system($main::postmap." -o $main::virtual_alias_maps") >> 8;
    die "$domain: postmap $main::virtual_alias_maps: $?" if $res != 0;
    print "postmaster\@$domain is $user\@$main::mydomain\nVirtual domain: $domain created\n";
}

sub domainDeletion {
# *****************************************
my ($domain, $user, $pwd, $style, $pw, $keepmail) = @_;
my @files;
my $domain1 = $domain; $domain1 =~ s/\./\\\./g;
    if (!$keepmail) { # used for conversion between virtual domain styles
        # delete /var/spool/virtual/domain
        print "Deleting file $main::virtual_mailbox_base/$domain/*\n";
        if (-e $main::virtual_mailbox_base."/$domain") {
            @files = <$main::virtual_mailbox_base/$domain/*>;
            if (@files) {
                unlink @files or die "Cannot delete files in $main::virtual_mailbox_base/$domain: $!";
            }
            rmdir $main::virtual_mailbox_base."/$domain" or die "Cannot remove $main::virtual_mailbox_base/$domain: $!";
        }

        # delete /etc/virtual/domain
        print "Deleting file $main::virtual_conf_base/$domain/*\n";
        if (-e $main::virtual_conf_base."/$domain") {
            @files = <$main::virtual_conf_base/$domain/*>;
            if (@files) {
                unlink @files or die "Cannot delete files in $main::virtual_conf_base/$domain: $!";
                rmdir $main::virtual_conf_base."/$domain" or die "Cannot remove $main::virtual_conf_base/$domain: $!";
            }
        }
    }

    # delete from /etc/postfix/virtual
    updateFile($domain, $main::virtual_alias_maps, "", '^([^@]*@)?'.$domain1.'\s');
    system($main::postmap." -o $main::virtual_alias_maps") == 0 or die "postmap $main::virtual_alias_maps: $?";

    if ($style eq 'jth') { # jth style virtual domains
        # delete from /etc/postfix/virtual_regexp
        my $tmp = $domain; $tmp =~ s/\./\\\\\\\./g;
        updateFile($domain, $main::virtual_regexp_maps, "", $tmp);

        # delete /home/user/.forward+*--domain
        # homedir ought to be fetched from the domain postmaster passwd entry ?
        print "Deleting file ".$pw->dir."/.forward+*--$domain\n";

        $tmp = $pw->dir."/.forward+*--$domain";
        @files = < $tmp >;
        if (@files) {
            unlink @files or die "Could not delete files ".$pw->dir."/.forward+*--$domain: $!";
        }

    } else { # postfix style virtual domains
        # delete from into /etc/postfix/transport
        updateFile($domain, $main::transport_maps, "", "^$domain1");

        #delete .forward+postmaster file like the .forward+postmaster--domain file
#       NO: user may administer other virtual domains

        if (!$keepmail) {
            # delete all mailboxes
            my $tmp ='grep -E "^[^@]+@'.$domain.' " '.$main::virtual_mailbox_maps;
            my @lin = `$tmp`;
            foreach my $l (@lin) {
                ($tmp, $tmp) = split(/\s+/, $l);
                if (-f "$main::virtual_mailbox_base/$tmp") {
                    print "Deleting file $main::virtual_mailbox_base/$tmp\n";
                    unlink "$main::virtual_mailbox_base/$tmp" or die "Cannot delete $main::virtual_mailbox_base/$tmp";
                }
            }
        }

        #delete all users in maps files
        my $filNew = "$main::virtual_mailbox_maps.new";
        if (-e $main::virtual_mailbox_maps) {
            print "Deleting line ^([^@]*@)?'.$domain from $main::virtual_mailbox_maps\n";
            system('grep -v -E "^([^@]*@)?'.$domain.' " '.$main::virtual_mailbox_maps.' >'.$filNew);
            system("mv $filNew $main::virtual_mailbox_maps") == 0 or die "mv $filNew $main::virtual_mailbox_maps: $?";
        }
        if (-e $main::virtual_uid_maps) {
            print "Deleting line ^([^@]*@)?'.$domain from $main::virtual_uid_maps\n";
            system('grep -v -E "^[^@]*@'.$domain.' " '.$main::virtual_uid_maps.' >'.$filNew);
            system("mv $filNew $main::virtual_uid_maps") == 0 or die "mv $filNew $main::virtual_uid_maps: $?";
        }
        if (-e $main::virtual_gid_maps) {
            print "Deleting line ^([^@]*@)?'.$domain from $main::virtual_gid_maps\n";
            system('grep -v -E "^[^@]*@'.$domain.' " '.$main::virtual_gid_maps.' >'.$filNew);
            system("mv $filNew $main::virtual_gid_maps") == 0 or die "mv $filNew $main::virtual_gid_maps: $?";
        }

    }
    print "Domain $domain (style $style) deleted\n";
    print "Mailboxes kept\n" if $keepmail;
    if ($style eq 'jth') {
#        print "Delete non-standard mailboxes (not in $main::virtual_mailbox_base/$domain) manually\n";
        print 'Update the '.$pw->dir.'/+forward file.'." Remove the $main::progid line if no more virtual domains for $user\n";
    }
}

sub updateFile {
# *****************************************
# Replace first line containing $searchexp with $line
# or add at end. Delete first found line if $line = ''
# Always deleting duplicates of $line
# Don't update/add if $keepold set to 'keepold'
# Delete postfix continue lines e.g. starting with whitespace, but keep empty lines
#
my ($domain, $fil, $line, $searchexp, $keepold) = @_;
my $filNew = "/tmp/vd$$.new";
my $found = 0;
my $none = 1;
my $res = '';
my $res1 = 1;
my $lin;
my $deletecontinue = 0;
$keepold = 'nokeepold' if !defined($keepold);
print "Inserting line '$line' in $fil\n" if $line ne '';
print "Deleting line '$searchexp' from $fil\n" if $line eq '';

if (-e $fil) {
    if (-r $fil) {
        $none = 0;
        open FIL, "<$fil" or $none = 2;
    } else {$none = 2;}
}
# print $searchexp."\n";

if (($none < 2) && (open FILN, ">$filNew")) {
    if ($none == 0) {
    while (($res eq '') && ($lin = <FIL>)) {
        chomp $lin; $lin =~ s/\s+$//;
        if (($found == 0) || ($line ne $lin)) {
        if ($lin =~ m!$searchexp!) {
            $res1 = print FILN "$line\n" if $line ne '' && $keepold ne 'keepold';
            syslog("info", "%s", "$filNew : $line") if $line ne '' && $keepold ne 'keepold';
            $found = 1;
            # set flag to delete continue lines
            $deletecontinue = 1;
        } else {
            if ($line eq '' || ($lin ne $line && !($deletecontinue && $lin =~ /^\s+[a-zA-Z0-9]+/))) {
                $res1 = print FILN "$lin\n";
                syslog("info", "%s", "$filNew: $lin");
                $deletecontinue = 0;
            }
        }
        $res = $! if !$res1;
        }
    }
    close FIL;
    }
    $res1 = print FILN "$line\n" if ($found == 0) && ($line ne '');
    $res = $! if !$res1;
    $res1 = close FILN;
    $res = $! if !$res1;
    if ($res eq '') {
        system("cp $filNew $fil") == 0 or $res = "Cannot copy $filNew to $fil. Keeping $filNew";
        if ($res eq '') {
            unlink $filNew or $res = "Cannot delete $filNew";
        }
    } else {
        unlink $filNew or $res = "Cannot delete $filNew";
    }
} else {
    $res = "ERROR: could not open $fil (RC: $none) or $filNew ($!)";
}
#syslog('info', "$domain: update:$fil: $line: $searchexp: $res");
if ($res ne '') {
  syslog('info', "$domain: update: f=$fil: l=$line: s=$searchexp: $res");
  print "$domain: $res\n";
}
return $res;
}

sub getUsers {
# *****************************************
# Utility to get all or a single user from the domain password file
# and the forwarding options for a single user as well as the domain style
# Input:  <domain name>, <postmaster password>, <optional virtual user name>
# Output: @main:users (with passwd file lines)
#         @main::forwards (with the single user forwarding lines)
#         $main::domainstyle
#         postmaster fields: $main::Pvuser,$main::Pupwd,$main::Puid,$main::Pgid, $main::Pname, $main::Phomedir
#     returns true if the password check is ok
#
my ($domain, $pwd, $user, $new) = @_;
my ($vuser, $upwd,$uid,$gid,$name,$mailbox);
$user = '' if !defined($user);
undef @main::users;
undef @main::forwards;
undef $main::Pvuser;

my $ll = $user.':';
#htmlPrint("$main::virtual_conf_base/$domain/passwd");
if (open PWD, "<$main::virtual_conf_base/$domain/passwd") {
    while (my $l = <PWD>) {
        chomp $l;
#htmlPrint("u=$l<br />"); # if $l =~ /^postmaster:/);
        ($vuser,$upwd,$uid,$gid,$name,$mailbox) = split(/:/,$l) if $l =~ /^$ll/;
        ($main::Pvuser,$main::Pupwd,$main::Puid,$main::Pgid, $main::Pname, $main::Phomedir) =
              split(/:/,$l) if $l =~ /^postmaster:/;
        push @main::users, $l if $user eq '' || $l =~ /^$ll/;
    }
    close PWD;
    my @tmp = sort @main::users; @main::users = @tmp; undef @tmp;
} else { # lookup real user
    my $pw = getpwnam($user);
    if ($pw) {
        $main::Phomedir = $pw->dir;
        $main::Pvuser = $pw->name;
        $main::Pname = $pw->name;
        $main::Pupwd = $pw->passwd;
        $main::Puid = $pw->uid;
        $main::Pgid = $pw->gid;
    } else {
        $main::Pvuser = $user;
        $main::Pname  = $user;
        $main::Pupwd = $pwd;
        $main::Phomedir = '/tmp';
        $main::Puid = 100;
        $main::Pgid = 100;
    }
}
if (defined($main::Phomedir)) {
    if (-e "$main::Phomedir/.forward+postmaster--$domain") {$main::domainstyle = 'jth'}
    else {$main::domainstyle = 'postfix'};
}

if ((!@main::users && defined($new)) || ($user eq '--newuser--')) {
    push @main::users, $user.'::'.$main::Puid.':'.$main::Pgid.'::.:';
} elsif (defined($vuser) && ($user ne '')) { # get users current forwarding options
    if ($main::domainstyle eq 'jth') {
        my $fwdfile = "$main::Phomedir/.forward+$user--$domain"; $fwdfile =~ s"//"/"g;
        if (open FWD, "<$fwdfile") {
            while (my $lin = <FWD>) {
            chomp $lin; $lin =~ s/\s+$//;
            push @main::forwards, $lin   if $lin !~ m!/$main::progid|/virtual(1)?!; # 'virtual' backwards compatible
        }
        close FWD;
    }
    } else { # postfix style virtual domain
    # get users possible entry in /etc/postfix/virtual
      my $lin = 'grep -E "^'.$user.'@'.$domain.'\s+" '.$main::virtual_alias_maps; $lin = `$lin`;
      $lin =~ s/\s+//g;
      if ($lin ne '') {
          ($lin,$lin) = split(/\s+/, $lin); # split off key
          $lin =~ s/\s+/ /g; @main::forwards = split(/\s+/, $lin);
      }
    }
}
#warn "Pu=$main::Puid Pg=$main::Pgid $main::virtualproc";
#foreach my $li (@main::users) { htmlPrint("get = $li<br />")}
return $main::Pvuser && (($pwd eq $main::Pupwd) || ($main::masterpw && $pwd && ($pwd eq $main::masterpw)));
}

sub displayLoginPage {
# *****************************************
# Display the login page for the user administration
#
my $domain = '';
$domain = $ENV{'HTTP_HOST'};
$domain =~ s/.+\.([-a-zA-Z0-9]+\.[-a-zA-Z0-9]+)$/$1/;
    $main::pagetitle = "Login VD-administration";
    htmlPrint("\n<center><form name=\"VIRTDOM\" method=\"$main::formmethod\" action=\"$main::nextform\" target=\"_top\">");
    htmlPrint("<input type=\"hidden\" name=\"trans\" value=\"LIST\" />");
    htmlPrint("<table summary=\"domain login\">");
    htmlPrint("<tr><td>$main::t[16]:</td><td><input type=\"text\" name=\"dom\" value=\"$domain\" /></td></tr>");
    htmlPrint("<tr><td>$main::t[11]:</td><td><input type=\"password\" name=\"pwd\" value=\"\" /></td></tr>");
    htmlPrint("<tr><td>&nbsp;</td><td><br /><input type=\"submit\" value=\"$main::t[47]\" style=\"background: yellow\" />");
    htmlPrint("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<input type=\"submit\" name=\"trans\" value=\"$main::t[48]\" style=\"background: green\" /></td></tr>");
    htmlPrint("</table><br /><br />&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;$main::t[46]: <input type=\"submit\" name=\"trans\" value=\"$main::t[49]\" /></form></center>");
}

sub displayUserHelp {
# *****************************************
# Display the help page for the user access to mail
#
my $domain = '';
$domain = $ENV{'HTTP_HOST'};
$domain =~ s/.+\.([-a-zA-Z0-9]+\.[-a-zA-Z0-9]+)$/$1/;

if (!$main::SMTPserver) {
    $main::SMTPserver = $main::t[44];
} elsif ($main::SMTPserver =~ /\.$/) {
    $main::SMTPserver =~ s/^(.*)\.$/$1/;
} else {
    $main::SMTPserver .= '.'.$domain;
}

if (!$main::POP3server) {
    $main::POP3server = $domain;
} elsif ($main::POP3server =~ /\.$/) {
    $main::POP3server =~ s/^(.*)\.$/$1/;
} else {
    $main::POP3server .= '.'.$domain;
}

if ($main::webmail) {
    if ($main::webmail =~ /\.$/) {
        $main::webmail =~ s/^(.*)\.$/$1/;
    } else {
        $main::webmail = "$domain/$main::webmail";
    }
}

    $main::pagetitle = "VD-administration help";
    htmlPrint("<table align=\"center\" summary=\"domain help\">");
    htmlPrint("<tr><td>$main::t[45]</td><td>$main::POP3server</td></tr>");
    htmlPrint("<tr><td>$main::t[40]</td><td>$main::t[41]$domain)</td></tr>");
    htmlPrint("<tr><td>$main::t[42]</td><td><a href=\"http://$main::webmail\">http://$main::webmail</a></td></tr>") if $main::webmail;
    htmlPrint("<tr><td>$main::t[43]</td><td>$main::SMTPserver</td></tr>");
    htmlPrint("</table>");
}


sub emailUpdateUser {
# *****************************************
# Submits a HTTP POST line to the postmaster@domain for updates
# and prints a result web page.
# The reason for this is, that the webserver does not have the rights to
# make updates to the admin user files. When mail is delivered by postfix
# it runs as the admin user
#
my ($domain, $pwd, $vuser, $style, $buffer) = @_;
    $main::pagetitle =$vuser;

    open UD, "|sh";
    print UD "$main::mailprog<<\"(END VD MAIL)\"\n";
    print UD "To: \"virtual domain admin\" <postmaster\@$domain>\n" if $style eq 'jth';;
    print UD "To: \"virtual domain admin\" <$main::postfixemail>\n" if $style eq 'postfix';
    print UD "From: \"virtual domain admin\" <$main::Pname\@$main::mydomain>\n" if $style eq 'jth';
    print UD "From: \"virtual domain admin\" <$main::postfixadmin\@$main::mydomain>\n" if $style eq 'postfix';
    print UD "Bcc: $main::bcc\n" if defined($main::bcc);
    print UD "Subject: *** Virtual domain update ***\n";
    print UD "MIME-Version: 1.0\n";
    print UD "Content-Type: text/plain; charset=ISO-8859-1\n";
    print UD "Content-Transfer-Encoding: 8bit\n\n";
    print UD $buffer; # as received from HTTP POST
    print UD "\n";
    print UD "(END VD MAIL)";
    close UD;

    htmlPrint("<center><br />*** $main::t[9]\n<br /><br />");
    htmlPrint("\n<form name=\"VIRTDOM\" method=\"$main::formmethod\" action=\"$main::nextform\" target=\"_top\">");
    htmlPrint("<input type=\"hidden\" name=\"trans\" value=\"LIST\" />");
    htmlPrint("<input type=\"hidden\" name=\"dom\" value=\"$domain\" />");
    htmlPrint("<input type=\"hidden\" name=\"pwd\" value=\"$pwd\" />");
    htmlPrint("<br /><input type=\"submit\" value=\" $main::t[8] \" style=\"background: yellow\" /></form></center>");
}

sub updateUser {
# *****************************************
# Performs the actual update of users in the virtual domain
#
my ($domain, $pwd, $vuser, $style, $upwd,$uid,$gid,$name,$mailbox, $action,
    $bounce, $keepold, $encrypt, $forward, @forwards) = @_;
$encrypt = 'no' if !$encrypt;
$keepold = 'no' if !$keepold;

syslog('info', "$domain: Update: $vuser mb=$mailbox action=$action ko=$keepold enc=$encrypt");

if (!getUsers($domain, $pwd, $vuser)) {
    syslog('info', "$domain: Password not OK. Exiting 67");
    exit 67;
}

my $fwdfil    = "$main::Phomedir/.forward+$vuser--$domain"; $fwdfil =~ s"//"/"g;
my $fwdfilP   = "$main::Phomedir/.forward"; $fwdfilP =~ s"//"/"g;
my $pwdfil    = "$main::virtual_conf_base/$domain/passwd";
my $pwdfilNew = $pwdfil.".new";
my $mbfil = '';
my $tmp = '';
my $res = '';

if ($mailbox eq '.') {$mbfil = $main::virtual_mailbox_base."/$domain/$vuser"}
else {$mbfil = $mailbox;}
$mbfil =~ s!//!/!g;
$action = 'update' if !defined($action);

$name = $vuser if !defined($name) || $name eq '';
$name =~ s/:/-/g;
$vuser =~ s/:/-/g;
$upwd =~ s/:/-/g;
$mailbox =~ s/:/-/g;

$tmp = "$vuser:$upwd:$uid:$gid:$name:$mailbox:" if $encrypt ne 'yes';
$tmp = "$vuser:".crypt($upwd, $upwd) if $encrypt eq 'yes' && length($upwd) != 13; # vm-pop3d
$tmp = "$vuser:".$upwd if $encrypt eq 'yes' && length($upwd) == 13; # vm-pop3d existing password

$tmp = '' if $action eq 'delete';

$res = updateFile($domain, $pwdfil, $tmp, "^$vuser:", $keepold); # keepold set when refreshing domain

if ($style eq 'postfix') {# postfix style virtual domains
    postfix_update($domain, $pwd, $vuser,$upwd,$uid,$gid,$name,$mailbox, $action, $bounce,
        $keepold, $encrypt, $forward, @forwards);
} else { # jth style virtual domains

if (($action eq 'delete') && ($vuser ne 'postmaster')) {
    # deleting the forwarding adresses
    if (-e $fwdfil) {
        unlink $fwdfil or die "$domain: Cannot delete $fwdfil: $!";
    }
    if (-e $mbfil && ! -d $mbfil) {
        unlink $mbfil  or die "$domain: Cannot delete $mbfil: $!";
    }
} else {
    # updating the forwarding file
    if ($vuser eq 'postmaster') {;
        $tmp = $main::virtualproc;
    } else {
        $tmp = $main::delivery;
    }

#    foreach my $f (@forwards) {
#        $f =~ s/\s+$//; $f = '"'.$f.'"' if $f =~ m!\||/!x; # command or file
#        $res = updateFile($domain, $fwdfil, $f, $f) if $main::allowcmd == 1 || ($f !~ m!\||/!);
#    }
    open FWD, ">$fwdfil.new" or die "Cannot open $fwdfil.new: $!";
    foreach my $f (@forwards) {
        $f =~ s/^\s*\|?"?([^"]*)"?\s*$/$1/;
#        $f = '"'.$f.'"' if $f =~ m!\||/!x; # command or file
        $f = '|"'.$f.'"' if $f !~ m!@!x; # command or file
       if ($main::allowcmd == 1 || ($f !~ m!\||/!)) {
          syslog("info", "%s", "$fwdfil: $f");
          $res = print FWD "$f\n";
       }
    }
    close FWD;
    system("mv -f $fwdfil.new $fwdfil") == 0 or die "Cannot move $fwdfil.new: $!";
    chmod 0644, $fwdfil;
    $res = updateFile($domain, $fwdfil, '|'.$tmp, $tmp) if !defined($forward) || $forward eq '';

    if ($vuser eq 'postmaster') {
        $bounce = (!defined($bounce) || $bounce eq '' ? '' : ' MAILBOUNCE');
        $tmp = $main::virtualproc; $tmp =~ s/(\.|\|)/\\$1/g;
        $res = updateFile($domain, $fwdfilP, '|"'.$main::virtualproc.$bounce.'"', $tmp);
    }
    if ($encrypt eq 'yes') { # make maildir and mailbox r/w for user $main::usermail as no uid/gid in passwd file
        my $pw = getpwnam($main::usermail) or die "No user $main::usermail";
        if (-e $mbfil ) {
            #chown $main::Puid, $pw->gid, $mbfil or die "Cannot change owner to $main::Puid.".$pw->gid." on $mbfil: $!";
            chmod 0660, $mbfil or die "Cannot change rights to 0660 on $mbfil: $!";
        }
        my $mbdir = substr($mbfil, 0, rindex('/', $mbfil));
        #chown $main::Puid, $pw->gid, $mbdir or die "Cannot change owner to $main::Puid.".$pw->gid." on $mbdir: $!";
        chmod 0770, $mbdir or die "Cannot change rights to 0770 on $mbdir: $!";
    }
}
}

}

sub listOneUser {
# *****************************************
# Generates a web page for a) a user b) a new user c) postmaster
#
my ($domain, $pwd, $user, $style) = @_;
my ($vuser, $upwd,$uid,$gid,$name,$mailbox, $forwardonly, $bounce, $encrypted);
my $fwdfilP = "$main::Phomedir/.forward"; $fwdfilP =~ s"//"/"g;
my $mailboxC = '';
$vuser = ""; $encrypted = '';
my $ll = $user.':';
foreach my $l (@main::users) {
#htmlPrint("$ll U=$l<br />");
    ($vuser,$upwd,$uid,$gid,$name,$mailbox) = split(/:/,$l) if $l =~ /^$ll/;
}

$encrypted = 'checked="checked"' if !defined($uid) && $upwd && length($upwd) == 13; # encrypted with crypt (vm-pop3d)
$name = $vuser if !defined($name);
$mailbox = "$domain/$user" if !defined($mailbox) || $mailbox eq '.';
$mailbox = $main::t[0] if $user eq '--newuser--';

my $roP = '';
$roP = 'readonly="readonly"' if $user eq 'postmaster';
my $ro1 = '';
$ro1 = 'readonly="readonly"' if $user ne '--newuser--';
$vuser = '' if $user eq '--newuser--';


if ($style eq 'jth') { # jth style virtual domains
    $mailbox = "$main::virtual_mailbox_base/".$mailbox if $mailbox !~ m!^/! && $user ne '--newuser--';
    $bounce = "fgrep MAILBOUNCE $main::Phomedir/.forward"; $bounce = `$bounce`; $bounce =~ s/\s+$//;
    $bounce =  (!defined($bounce) || ($bounce eq '') ? 'no' : 'yes');

    $forwardonly = "$main::Phomedir/.forward+$user--$domain";
    if (-e $forwardonly) { # if command to store in mailbox is not found then forwarding only
        $forwardonly = 'grep -E "'."/$main::progid|/virtual(1)?".'" '.$forwardonly;
        $forwardonly = `$forwardonly`; $forwardonly =~ s/\s+$//;
        $forwardonly =  ($forwardonly eq '' ? 'yes' : 'no');
     } else {$forwardonly = 'no';}
} else {
    $mailboxC = "$main::virtual_mailbox_base/" if ($user ne 'postmaster');
    $bounce = $main::postmap.' -q @'.$domain." $main::virtual_alias_maps"; $bounce = `$bounce`; $bounce =~ s/\s+$//;
    $bounce =  (!defined($bounce) || ($bounce eq '') ? 'yes' : 'no');

    $forwardonly = $main::postmap.' -q '."$user\@$domain $main::virtual_alias_maps";
    $forwardonly = `$forwardonly`; $forwardonly =~ s/\s+/ /; $forwardonly =~ s/\s+$//;
    @main::forwards = split(/\s+/,$forwardonly);

    if (@main::forwards && $main::forwards[0] eq "$user\@$domain" && $user ne 'postmaster') {
        shift @main::forwards;
        $forwardonly = 'no';
    } else { $forwardonly = 'yes' }

    if (!$uid) {
         $uid = $main::postmap." -q $user\@$domain $main::virtual_uid_maps"; $uid = `$uid`; $uid =~ s/\s+$//;
    }
    if (!$gid) {
         $gid = $main::postmap." -q $user\@$domain $main::virtual_gid_maps"; $gid = `$gid`; $gid =~ s/\s+$//;
    }
}
    $uid = $main::Puid if !$uid;
    $gid = $main::Pgid if !$gid;
    use User::grent;
    my $pw = getpwuid($uid);
    my $uidN = ($pw ? $pw->name : '???');
       $pw = getgrgid($gid);
    my $gidN = ($pw ? $pw->name : '???');

    $bounce      =  ($bounce eq 'yes' ? 'checked="checked"' : '');
    $forwardonly =  ($forwardonly eq 'yes' ? 'checked="checked"' : '');
    my $email = '&lt;'.($style eq 'postfix' ? "$main::postfixemail" : "postmaster\@$domain").'&gt;';

    $main::pagetitle ="$main::t[24] $user";
    htmlPrint("<form name=\"VIRTDOM\" method=\"$main::formmethod\" action=\"$main::nextform\" target=\"_top\">");
    htmlPrint("<input type=\"hidden\" name=\"trans\" value=\"UPDATE\" />");
    htmlPrint("<input type=\"hidden\" name=\"dom\" value=\"$domain\" />");
    htmlPrint("<input type=\"hidden\" name=\"pwd\" value=\"$pwd\" />");
    htmlPrint("<input type=\"hidden\" name=\"orguser\" value=\"$user\" />");
    htmlPrint("<input type=\"hidden\" name=\"lang\" value=\"$main::lang\" />");
    if ($main::showmailbox ne 'yes') {
        htmlPrint("<input type=\"hidden\" name=\"mailbox\" value=\"$mailbox\" />");
    }
    if ($main::showuidgid ne 'yes') {
        htmlPrint("<input type=\"hidden\" name=\"uid\" value=\"$uid\" />");
        htmlPrint("<input type=\"hidden\" name=\"gid\" value=\"$gid\" />");
    }
    htmlPrint("<table align=\"center\" summary=\"user data\">");
    htmlPrint("<tr><td align=\"right\">$main::t[3]:</td><td><input type=\"text\" name=\"user\" value=\"$vuser\" $ro1 /> \@$domain</td></tr>");
    htmlPrint("<tr><td align=\"right\">$main::t[4]:</td><td> <input type=\"text\" name=\"name\" value=\"$name\" $roP /></td></tr>");
    htmlPrint("<tr><td align=\"right\">$main::t[11]:</td><td>  <input type=\"text\" name=\"upwd\" value=\"$upwd\" />");
    htmlPrint("<input type=\"checkbox\" name=\"encrypt\" value=\"yes\" $encrypted /> $main::t[33]") if $main::vm_pop3d eq 'yes';
    htmlPrint("</td></tr>");

    if ($main::showmailbox eq 'yes') {
        htmlPrint("<tr><td align=\"right\">".($user eq 'postmaster' ?$main::t[28]:$main::t[5]).":</td><td>
                   $mailboxC <input type=\"text\" name=\"mailbox\" value=\"$mailbox\" size=\"64\" $roP /></td></tr>");
    }
    if ($main::showuidgid eq 'yes') {
        htmlPrint("<tr><td align=\"right\">$main::t[14]:</td><td> <input type=\"text\" name=\"uid\" value=\"$uid\" size=\"6\" $roP /> $uidN</td></tr>");
        htmlPrint("<tr><td align=\"right\">$main::t[15]:</td><td> <input type=\"text\" name=\"gid\" value=\"$gid\" size=\"6\" $roP /> $gidN</td></tr>");
    }
    my $i = 0;
    my $tmp = ($style eq 'postfix' ? $main::t[12] : $main::t[34]);
    foreach my $f (@main::forwards) {
        $f =~ s/\s+$//; $i++;
        if ($f) {
            $f =~ s/"/&quot;/g; $f =~ s/\|//;
            htmlPrint("<tr><td align=\"right\">$tmp $i:</td><td> <input type=\"text\" name=\"fwd$i\" value=\"$f\" size=\"64\" /></td></tr>")
        }
    }
    while ($i < 2) {
        $i++;
        htmlPrint("<tr><td align=\"right\">$tmp $i:</td><td> <input type=\"text\" name=\"fwd$i\" value=\"\" size=\"64\" /></td></tr>");
    }
    $i++;
    htmlPrint("<tr><td align=\"right\">$tmp $i:</td><td> <input type=\"text\" name=\"fwd$i\" value=\"\" size=\"64\" /></td></tr>");
    htmlPrint("<tr><td align=\"right\">$main::t[19]:</td><td> <input type=\"checkbox\" name=\"bounce\" value=\"yes\" $bounce /></td></tr>") if $user eq 'postmaster';
    htmlPrint("<tr><td align=\"right\">$main::t[27]:</td><td> <input type=\"text\" name=\"style\" value=\"$style\" size=\"10\" readonly=\"readonly\" /></td></tr>") if $user eq 'postmaster';

    htmlPrint("<tr><td align=\"right\">$main::t[20]:</td><td> <input type=\"checkbox\" name=\"forwardonly\" value=\"yes\" $forwardonly /></td></tr>") if $user ne 'postmaster';
    htmlPrint("<tr><td align=\"right\">$main::t[13]:</td><td> <input type=\"checkbox\" name=\"action\" value=\"delete\" /></td></tr>") if $user ne 'postmaster' && $ro1 ne '';
    htmlPrint("</table><br /><center><input type=\"submit\" value=\" $main::t[10] \" style=\"background: yellow\" />");
    htmlPrint("</center></form>");
    htmlPrint("<p><font size=\"-1\">");
    htmlPrint("$main::t[30] $email<br />") if $user eq 'postmaster';
    htmlPrint("$main::t[26]<br />") if $user ne 'postmaster' && $main::showmailbox eq 'yes';
    htmlPrint("$main::t[29]<br />") if $style ne 'postfix' && $main::showuidgid eq 'yes';
    htmlPrint("</font></p>");
}

sub listUsers {
# *****************************************
# Generates a web page displaying all users and option to add a new user
#
my ($domain, $pwd, $style) = @_;

my $tel = 0;

    $main::pagetitle = "$main::t[25] $domain";

    htmlPrint("\n<form name=\"VIRTDOM\" method=\"$main::formmethod\" action=\"$main::nextform\" target=\"_top\">");
    htmlPrint("<input type=\"hidden\" name=\"trans\" value=\"LIST1\" />");
    htmlPrint("<input type=\"hidden\" name=\"dom\" value=\"$domain\" />");
    htmlPrint("<input type=\"hidden\" name=\"pwd\" value=\"$pwd\" />");
    htmlPrint("<table border=\"1\" align=\"center\" summary=\"$main::t[25] $domain\">");
    htmlPrint("<tr><th>&nbsp;</th><th>$main::t[2]</th><th>$main::t[3]</th><th>$main::t[4]</th><th>$main::t[5]</th><th>$main::t[21]</th></tr>");
    foreach my $l (@main::users) {
        my ($user,$pwd,$uid,$gid,$name,$mailbox) = split(/:/,$l);
        $name = $user if !defined($name);
        $mailbox = $main::t[0] if !defined($mailbox) || $mailbox eq '.';
        my $anymail = $main::t[23];

        my $mbfil = (lc($mailbox) eq $main::t[0] ? "$domain/$user" : $mailbox);

        if ($user eq 'postmaster') { # get postmaster mailbox
            my $pw = getpwnam($name);
            if ($pw) {
                my $maildir = $pw->dir."/$main::home_mailbox";
                $maildir =~ s!(^.*)/[^/]*$!$1!;
                if (($main::home_mailbox ne '') && -d $maildir) {
                    $mbfil = $pw->dir."/$main::home_mailbox";
                } elsif (-d $main::mail_spool_directory) {
                    $mbfil = $main::mail_spool_directory."/$name";
                } elsif (-d "/var/mail") {
                    $mbfil = "/var/mail/$name";
                }
            } elsif ($main::home_mailbox ne '') {
                $mbfil = "$mailbox/$main::home_mailbox";
            } else {
                $mbfil = "$main::mail_spool_directory/$name";
            }
        } elsif ($style eq 'postfix' || $mailbox !~ m!^/!) {
            $mbfil = $main::virtual_mailbox_base.'/'.$mbfil;
        }
        my $fsize = (-s $mbfil); if (defined($fsize)) {$fsize = $fsize / 1024; } else {$fsize = 0}
        $fsize = sprintf("%.1f", $fsize);
        $anymail = $main::t[22] if $fsize > 0;

        $tel++;
        htmlPrint("<tr><td align=\"right\">$tel</td><td><input type=\"radio\" name=\"user\" value=\"$user\" /></td>");
        htmlPrint("<td>$user</td><td>$name</td><td>$mailbox</td><td align=\"right\">$fsize kB</td></tr>");
    }
    htmlPrint("<tr><td></td><td><input type=\"radio\" name=\"user\" value=\"--newuser--\" /></td>");
    htmlPrint("<td>$main::t[1]</td><td></td><td></td></tr>");
    htmlPrint("</table><br /><center><input type=\"submit\" value=\" $main::t[6] \" style=\"background: yellow\" /></center></form>");
}

sub htmlHeader {
# *****************************************
# Prints the HTML-header for all web pages
#
my ($title, $domain) = @_;
return if $main::command !~ /^WEB/ || $main::headerwritten;
$title = $main::pagetitle if !$title;
$main::headerwritten = 1;

$domain = '' if !defined($domain);
htmlPrint('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"');
htmlPrint('    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">');
htmlPrint('<html xmlns="http://www.w3.org/1999/xhtml">');
htmlPrint('<head>');
htmlPrint('<meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1" />');
htmlPrint('<meta http-equiv="Content-Style-Type" content="text/css" />');
htmlPrint('<meta http-equiv="Content-Script-Type" content="text/javascript" />');
htmlPrint('<style type="text/css"> <!-- ');
htmlPrint("body.vd {font-family: Verdana, Arial, Helvetica, sans-serif; font-size: 12pt; background: $main::bg; color: $main::fg;}");
htmlPrint("a:link {color: $main::link} a:visited {color: $main::visited} a:active {color: #CC0000} ");
htmlPrint(".cr { font-size: 60%; }");
htmlPrint('--></style>');
htmlPrint("<title>$title</title>");
htmlPrint('</head><body class="vd">');
htmlPrint("$main::banner") if $main::banner;
htmlPrint("<center><h2>$main::t[7] $domain</h2></center>");
}

sub htmlFooter {
# *****************************************
# Prints the HTML-footer for all web pages
#
return if $main::command !~ /^WEB/ || !$main::headerwritten;
#htmlPrint('<p><a href="http://validator.w3.org/check/referer"><img src="http://www.w3.org/Icons/valid-xhtml10"');
htmlPrint('<p><a href="http://validator.w3.org/check/referer"><img src="/icons/valid-xhtml10"');
htmlPrint('alt="Valid XHTML 1.0!" height="31" width="88" border="0" /></a>');
htmlPrint("<span class=\"cr\">$main::progname $main::progversion &copy; 2002 J. Thomsen</span>");
htmlPrint('</p></body></html>');
$main::headerwritten = 0;
}

sub htmlPrint {
# *****************************************
# Conditional print of HTML-line. Nothing printed if not called from web server
#
my ($str) = @_;
    if ($main::command !~ /^WEB/) {
        $str =~ s/<[^>]+>//g;
        $str =~ s/&lt;/</g;
        $str =~ s/&gt;/>/g;
    } else {
        htmlHeader($main::pagetitle, $main::domain) if !$main::headerwritten;
    }
    print $str."\n" if $str ne '';
}

sub texts {
# *****************************************
# Texts for the web interface. Commandline is always English
#
my ($lang) = @_;
my @t;
if ($lang eq 'da') { # Danish
@t = (
'[standard]',
'Ny bruger',
'Ret',
'Email',
'Navn',
'Postkasse',
'Send', # button listUsers
'Brugeradministration',
'Tilbage',
'Opdatering af bruger startet. Vent lidt og tryk s p knappen nedenfor.',
'Opdater', # button listOneUser
'Kodeord', # 11
'Videresend',
'Slet bruger',
'Brugernr',
'Gruppenr',
'Domne',
'Ukendt kommando',
'Forkert kodeord',
'Afvis post fra ukendte brugere',
'Kun videresendelse<br />(ingen lokal postkasse)', # 20
'Post',
'Ja',
'Nej',
'Bruger',
'Brugere i',
'OBS. Ved brug af en ikke-standard postkasse, s vr sikker p, at din software kan lse den, nr post skal hentes. Se <a href="http://jth.net/virtual">GNU pop3d</a>',
'Domnetype',
'Admin homedir',
'Brugernr og gruppenr bruges ikke for denne type domne, men kan vre nyttige ved konvertering til et postfix type domne',
'Opdatering sendes til', #30
'Systemadministratoren skal checke disse parametre for korrekte vrdier',
'Opret (konverter/genopret) og slet virtuelle domner fra en kommandolinje logget in som root',
'Krypteret (kun vm-pop3d)',
'Videresend-adr. eller "kommando"',
'er ikke i korrekt form til brug for',
'skulle indeholde linjer som',
'OK hvis tom',
'ikke fundet',
'Lav tomme "hash:" databaser p flgende mde',
'POP3 login-navn', # 40
'din email-adresse (fx xxx@',
'Adgang til post via browser',
'SMTP server',
'brug din udbyders server',
'POP3 server',
'Kun for systemadministrator',
'Login', # button displayLoginPage
'Info',  # button displayLoginPage
'Systemcheck',  # button displayLoginPage
);
} else { # English
@t = (
'[default]',
'New user',
'Edit',
'Email',
'Name',
'Mailbox',
'Submit',  # button listUsers
'User administration',
'Back',
'Update of user started. Please, wait a little while, then push the button below',
'Update', # button listOneUser
'Password', # 11
'Forward',
'Delete user',
'User id',
'Group id',
'Domain',
'Unknown command',
'Wrong password',
'Bounce mail from unknown users',
'Only forwarding<br />(no local mailbox)', # 20
'Mail',
'Yes',
'No',
'User',
'Users in',
'OBS. If using a non-standard mailbox, make sure your software can access it, when retrieving the mail e.g. See <a href="http://jth.net/virtual">GNU pop3d</a>',
'Domain style',
'Admin homedir',
'User id and group id are not used for this style of domain, but could be useful if converting to a postfix style domain',
'Updates are mailed to', # 30
'OBS. The system administrator must check these parameters for correct values',
'Create (convert/recreate) and delete virtual domains from a commandline logged in as root',
'Encrypted (vm-pop3d only)',
'Fwd. address or "command"',
'is not in proper form for use by',
'should contain lines like',
'OK if empty',
'not found',
'Make empty "hash:" databases this way',
'POP3 login name', # 40
'your email address (e.g. xxx@',
'Mail access by browser',
'SMTP server',
'use the one provided by your ISP',
'POP3 server',
'System administrators only',
'Login', # button displayLoginPage
'Info',  # button displayLoginPage
'Systemcheck',  # button displayLoginPage
);
}
return @t;
}


sub postfix_get_var {
# *****************************************
# Read a parameter from /etc/postfix/main.cf
#
my ($key, $all, $default) = @_;
my $val;
my $dummy;
my @fil;
my $def='';

if (!@main::maincf) {
#   @main::maincf = `grep -v -E "^#" /etc/postfix/main.cf`;
    @main::maincf = `/usr/sbin/postconf`;
}

my @vals = grep(/^$key\s+=/, @main::maincf); $val = (@vals ? $vals[0] : '');
#my $getcontinue = 0;
#foreach my $lin (@main::maincf) {
#    if ($lin =~ /^$key\s+=/) {
#        $lin =~ s/\s+/ /g;
#        $val = $lin;
#        $getcontinue = 1;
#    } elsif ($getcontinue) {
#        if ($lin =~ /^\s+[a-zA-Z0-9]+/) {
#            $lin =~ s/\s+/ /g;
#            $val .= ' '.$lin;
#        } else { $getcontinue = 0 }
#    }
#}

($dummy, $val) = split(/=/, $val) if $val =~ /=/;
$val =~ s/\s+//; $val =~ s/\s+$//;
if (!$all) {
    ($dummy, $val) = split(/:/, $val) if $val =~ /:/; # discard eg. hash:...
    ($val, $dummy) = split(/,/, $val) if $val =~ /,/; # only first value.
    ($val, $dummy) = split(/\s+/, $val); # only first value.
}
my $obs = '';
my $prefix = '';
($prefix, $dummy) = split(/:/, $default) if $default && (index($default, ':') > 0);
$prefix .= ':' if $prefix;

if (!$val) {
    $obs = '** OBS ** '; $val = '';
    $def = "</td><td>[ $default ]" if $default;
} else {
    $prefix = '' if $val =~ /$prefix/;
}

htmlPrint("<tr><td>$obs"."main.cf: $key </td><td>=</td><td> $prefix$val$def</td></tr>")
    if ($main::command =~ /^(WEBCHECK)?$/) && (!$all || ($all && $all !~ /noprint/));

return $val;
}

sub postfix_update {
# *****************************************
# Performs the actual update of a user in the virtual domain
#
my ($domain, $pwd, $vuser,$upwd,$uid,$gid,$name, $mailbox, $action, $bounce, $keepold, $encrypt, $forwardonly, @forwards) = @_;
my ($domain1, $tmp);
my $vm = 0;
my $res = '';
$mailbox = "$domain/$vuser" if $mailbox eq '.';
$domain1 = $domain; $domain1 =~ s/\./\\\./g;

if ($action eq 'delete') {
    return if $vuser eq 'postmaster'; # only to be deleted with entire domain
    # delete users mailbox
    my $tmp = $main::postmap." -q $vuser\@$domain $main::virtual_mailbox_maps"; $tmp = `$tmp`; $tmp =~ s/\s+$//;
#    my $tmp ='grep -E "^'.$vuser.'@'.$domain1.' "'.$main::virtual_mailbox_maps;
#    my @lin = `$tmp`;
#    foreach my $l (@lin) {
#      ($tmp, $tmp) = split(/\s+/, $l);
       if (-f "$main::virtual_mailbox_base/$tmp") {
            unlink "$main::virtual_mailbox_base/$tmp" or die "$domain: Cannot delete $main::virtual_mailbox_base/$tmp";
            syslog('info', "$domain: Deleted $main::virtual_mailbox_base/$tmp\n");
       }
#    }

    # delete in /etc/postfix/virtual
    $res = updateFile($domain, $main::virtual_alias_maps, '', "^$vuser\@$domain1\\s+"); $vm = 1;

    $res = updateFile($domain, $main::virtual_mailbox_maps, "", "^$vuser\@$domain1\\s+");
    $res = updateFile($domain, $main::virtual_uid_maps, "", "^$vuser\@$domain1\\s+");
    $res = updateFile($domain, $main::virtual_gid_maps, "", "^$vuser\@$domain1\\s+");
} else {
    if ($vuser ne 'postmaster') {
        $res = updateFile($domain, $main::virtual_mailbox_maps, "$vuser\@$domain $mailbox", "^$vuser\@$domain1\\s+");
        $res = updateFile($domain, $main::virtual_uid_maps, "$vuser\@$domain $uid", "^$vuser\@$domain1\\s+");
        $res = updateFile($domain, $main::virtual_gid_maps, "$vuser\@$domain $gid", "^$vuser\@$domain1\\s+");
        if ($encrypt eq 'yes') { # make maildir and mailbox r/w for user $main::usermail as no uid/gid in passwd file
            my $mbfil = "$main::virtual_mailbox_base/$mailbox";
            my $pw = getpwnam($main::usermail) or die "No user $main::usermail";
            if (-e $mbfil ) {
                $tmp = "chown ".$uid.':'.$pw->uid." $mbfil"; $tmp = `$tmp`; syslog('info', $tmp);

                #chown $uid, $pw->gid, $mbfil or die "Cannot change owner to $uid.".$pw->gid." on $mbfil: $!";
                chmod 0660, $mbfil or die "Cannot change rights to 0660 on $mbfil: $!";
            }
            my $mbdir = substr($mbfil, 0, rindex('/', $mbfil));
            #chown $uid, $pw->gid, $mbdir or die "Cannot change owner to $uid.".$pw->gid." on $mbdir: $!";
            chmod 0770, $mbdir or die "Cannot change rights to 0770 on $mbdir: $!";
        }
    }

    # if a catch-all for the domain is defined a 'user@domain user@domain' must be defined, so always insert it

    $tmp = "$vuser\@$domain";

    if (@forwards) {
        $tmp .= " $vuser\@$domain" if (!defined($forwardonly) || $forwardonly ne 'yes') && ($vuser ne 'postmaster');
        foreach my $f (@forwards) {
            $tmp .= " $f" if $f !~ m!\||/!; # no delivery to commands or files
#            syslog('info', "$domain: bruger $f");
        }
    } else {$tmp .= " $vuser\@$domain";}
    # insert into /etc/postfix/virtual
    $res = updateFile($domain, $main::virtual_alias_maps, $tmp, "^$vuser\@$domain1\\s+"); $vm = 1;

    if ($vuser eq 'postmaster') {
        $bounce = '' if !defined($bounce);
        if ($bounce ne 'yes') {
            $tmp = "\@domain postmaster\@$domain";
        } else { $tmp = ''; }
        # insert/delete in /etc/postfix/virtual
        $res = updateFile($domain, $main::virtual_alias_maps, $tmp, "^\@$domain1\\s+"); $vm = 1;
    }

}
if ($vm == 1) {
    $res = system($main::postmap." -o $main::virtual_alias_maps") >> 0;
    die "$domain: postmap $main::virtual_alias_maps failed: $?" if $res != 0;
}
$res = system($main::postmap." -o $main::virtual_mailbox_maps") >> 8;
die "$domain: postmap $main::virtual_mailbox_maps failed: $?" if $res != 0;
$res = system($main::postmap." -o $main::virtual_uid_maps") >> 8;
die "$domain: postmap $main::virtual_uid_maps failed: $?" if $res != 0;
$res = system($main::postmap." -o $main::virtual_gid_maps") >> 8;
die "$domain: postmap $main::virtual_gid_maps failed: $?" if $res != 0;
}
